# -*- coding: utf-8 -*-
import os
import argparse
import logging
import sys
from bilibili_toolman import providers, __version__, __desc__

class AttribuitedDict(dict):
    def __getattr__(self, name):
        return self[name]


pbar = None
global_args = {
    "cookies": {
        "help": "登陆： 使用 Cookies 登陆，即使用 Web API （不可多 P 上传） ( 需要 SESSDATA 及 bili_jct ) e.g.SESSDATA=cb0..; bili_jct=6750..."
    },
    "sms": {
        "help": "登陆：使用短信验证码登陆，即使用 上传助手 API （可多 P 上传）（需手动交互）（有日获取限制，请配合 --save 使用）",
        "default": False,
        "action": "store_true",
    },
    "load": {"help": "登陆：加载凭据，同时用于上传及投稿"},
    "load_upload": {"help": "登陆：使用该凭据上传，而不用--load凭据上传"},
    "load_submit": {"help": "登陆：使用该凭据投稿，而不用--load凭据投稿"},
    "save": {
        "help": "登陆：向stdout输出当前登陆凭据并退出（其他输出转移至stderr）",
        "default": False,
        "action": "store_true",
    },
    "http": {"help": "强制使用 HTTP （不推荐）", "default": False, "action": "store_true"},
    "noenv": {"help": "上传时，不采用环境变量（如代理）", "default": False, "action": "store_true"},
    "cdn": {
        "help": "上传用 CDN （限 Web API) （对应 网宿（适合海外），七牛，百度（默认），七牛，谷歌，百度）",
        "choices": ["ws", "qn", "bda2", "kodo", "gcs", "bos"],
        "default": "bda2",
    },
}
local_args = {
    "opts": {"help": "解析可选参数 ，详见 --opts 格式", "default": ""},
    "thread_id": {"help": "分区 ID", "default": 17},
    "tags": {"help": "标签", "default": "转载"},
    "desc": {"help": '描述格式 e.g. "原描述：{desc}" (其他变量详见下文)（仅稿件有描述）', "default": "{desc}"},
    "title": {
        "help": '标题格式 e.g. "[Youtube] {title} (其他变量详见下文)（使用于稿件及分P）"',
        "default": "{title}",
    },
    "seperate_parts": {
        "help": "不分P （e.g. --youtube [播放列表],--localfile [文件夹]）独立投稿（不分P）（Web上传默认不分 P）",
        "default": False,
        "action": "store_true",
    },
    "no_upload": {"help": "只下载资源", "default": False, "action": "store_true"},
    "no_submit": {
        "help": "不提交稿件，适用于获取filename参数",
        "default": False,
        "action": "store_true",
    },
    "original": {"help": "设置稿件为原创", "default": False, "action": "store_true"},
    "no_reprint": {"help": "设置稿件不允许转载", "default": False, "action": "store_true"},
}
arg_epilog = """
变量：
    {title},{desc} 等变量适用于：
        --title, --desc, --tags

通用变量:
    {title}     -       原标题
    {desc}      -       原描述
    {safe_korean_title}
                -       【韩文】替换韩文为特殊字符的标题
    {roma_korean_title}
                -       【韩文】替换韩文为罗马音的标题 (需要安装 korean_romanizer)
本工具支持将给定视频源转载至哔哩哔哩

详见项目 README 以获取更多例程 ： github.com/greats3an/bilibili-toolman
"""


def setup_logging():
    from tqdm.std import tqdm as tqdm_c

    class SemaphoreStdout:
        @staticmethod
        def write(__s):
            # Blocks tqdm's output until write on this stream is done
            # Solves cases where progress bars gets re-rendered when logs
            # spews out too fast
            with tqdm_c.external_write_mode(file=sys.stdout, nolock=False):
                return sys.stdout.write(__s)

    import coloredlogs

    coloredlogs.DEFAULT_LOG_FORMAT = (
        "[ %(asctime)s %(name)8s %(levelname)6s ] %(message)s"
    )
    coloredlogs.install(level=logging.DEBUG, stream=SemaphoreStdout, isatty=True)
    logging.getLogger("urllib3").setLevel(logging.CRITICAL)
    logging.getLogger("PIL.Image").setLevel(logging.CRITICAL)    

def truncate(string, max):
    """truncate & add ... to string over the length of `max`"""
    if len(string) > max:
        string = string[: max - 3] + "..."
    return string

def prepare_temp(temp_path: str):
    if not os.path.isdir(temp_path):
        os.mkdir(temp_path)
    os.chdir(temp_path)
    return True


def report_progress(current, max_val):
    from bilibili_toolman.cli import precentage_progress

    precentage_progress.report(current, max_val)


def _enumerate_providers():
    provider_dict = dict()
    for provider in dir(providers):
        if not "provider_" in provider:
            continue
        provider_name = provider.replace("provider_", "")
        provider_dict[provider_name] = getattr(providers, provider)
    return provider_dict


provider_args = _enumerate_providers()


def _create_argparser():
    p = argparse.ArgumentParser(
        description="%s %s 使用帮助" % (__desc__,__version__),
        formatter_class=argparse.RawTextHelpFormatter,
        epilog=arg_epilog,
    )
    g = p.add_argument_group("身份设置 （随方式优先级排序）")
    for arg_key, arg in global_args.items():
        g.add_argument("--%s" % arg_key, **arg)
    # global args
    g = p.add_argument_group("上传设置")
    for arg_key, arg in local_args.items():
        g.add_argument("--%s" % arg_key, **arg)
    # local args (per source)
    g = p.add_argument_group(
        '解析可选参数 "opts" （格式 ： [参数1]=[值1];[参数2]=[值2] (query-string)）\n'+
        '视频源参数'
    )
    for provider_name, provider in provider_args.items():
        g.add_argument(
            "--%s" % provider_name,
            metavar="%s-URL" % provider_name.upper(),
            type=str,
            help="%s\n   参数:%s" % (provider.__desc__, provider.__cfg_help__),
        )
    return p


def prase_args(args: list):
    if len(args) < 2:
        parser = _create_argparser()
        parser.print_help()
        return
    args.pop(0)  # remove filename
    parser = _create_argparser()
    global_args_dict = AttribuitedDict()
    for k, v in parser.parse_args(args).__dict__.items():
        if k in global_args:
            global_args_dict[k] = v
            if not "--%s" % k in args:
                continue
        elif k in local_args and local_args[k]["default"] != v:
            # raise argparse.ArgumentError(None ,'"上传设置" (--%s) 仅应该出现于视频源后 (e.g. --youtube "..." --tags ..) ' % k)
            pass # TODO : Properly implement this check
    """pre-parse : fetch global args,then remove them"""
    local_args_group = []
    current_line = []
    current_provider = ""

    def add():
        args = parser.parse_args(current_line).__dict__
        args["resource"] = args[current_provider]
        local_args_group.append(
            (provider_args[current_provider], AttribuitedDict(args))
        )

    i = 0
    while i < len(args):
        if args[i][2:] in provider_args:
            # a new proivder. terminates till next provider
            current_provider = args[i][2:]
            current_line = ["--" + current_provider]
            for i in range(i + 1, len(args)):
                if args[i][2:] in provider_args:
                    break
                current_line.append(args[i])
            add()
            i -= 1
        i += 1
    return global_args_dict, local_args_group



'''Languages aren't created equally to begin with. And some just isn't liked...by Bilibili.

Sanitizers are attempts to address this issue. They work by replacing / translating
"illicit" words / phrases into "better, more suitable" forms.

A sanitizer function takes 
* 1 positional argument (string)
* any number of keyword arguments

And returns a tuple
* string to be matched by regex
* regex to be used 
* replacement strategey

... where a `replacement strategey` is a pure function.
* takes one string input
* give one string output

Sanitizers are constructed with `@sanitizer` wrapper. You'll find some examples in the code below.
'''
from functools import wraps
def sanitizer(function):
    @wraps(function)
    def wrapper(string,**kw):
        import re
        string, regex, replace_strategy = function(string, **kw)
        regex = re.compile(regex)
        index, output = 0, ''    
        for result in regex.finditer(string):
            begin,end = result.span()
            output += string[index:begin]
            output += replace_strategy(string[begin:end])
            index = end
        output += string[index:]
        return output
    return wrapper

def sanitizer_strategy_replace_with(char):
    '''Replaces forbidden characters with other one(s)'''
    def strategy(str):
        return char * len(str)
    return strategy

@sanitizer
def sanitize_korean(string,replace_with="█",replace_with_romaji=False):  
    '''Korean is banned by the site...huh

    Switches:
        replace_with_romaji (bool) : If True, `korean-romanizer` will be used to translate Hangul into romal forms. (Defaults to False).
        replace_with (str) : Replaces korean characters with something else (Defaults to █).
    '''    
    strategy = sanitizer_strategy_replace_with(replace_with)
    if replace_with_romaji:
        def korean_romanization(str):
            # Implements Korean romanization : https://github.com/osori/korean-romanizer
            from korean_romanizer.romanizer import Romanizer
            return Romanizer(str).romanize()
        strategy = korean_romanization
    return string, '[\uAC00-\uD7A3]+' , strategy



# -*- coding: utf-8 -*-
from json import JSONDecodeError
import os
from functools import wraps
from threading import Lock
from typing import Tuple
from requests import Session
from io import IOBase
from queue import Queue
import time

from requests.models import Response

# region Wrappers
def JSONResponse(classfunc) -> dict:
    """Decodes `Response`s content to JSON dict"""
    @wraps(classfunc)
    def wrapper(session: Session, *args, **kwargs):
        try:
            response: Response = classfunc(session, *args, **kwargs)
            return response.json()
        except JSONDecodeError:
            assert response.status_code == 200, 'HTTP %s\n%s\n%s' % (response.status_code,response.request.url,response.text)
            return response.text
    return wrapper


# endregion

# common models
class LoginException(Exception):
    def __init__(self, resp: Response, desc="") -> None:
        self.resp = resp
        self.desc = desc
        super().__init__(
            "登陆失败：%s\n<Response %s>:%s" % (desc, resp.status_code, resp.text)
        )


class ReprExDict(dict):
    """fancy __repr__-ed dict"""

    def __repr__(self) -> str:
        return "(%s)" % " ".join(["%s/%s" % (k, v) for k, v in self.items()])


class FileManager(dict):
    """single-instance threadsafe file IO manager"""

    CHUNK_SIZE = 2**16

    def __init__(self) -> None:
        super().__init__()
        self.lock = Lock()

    def open(self, path):
        self.lock.acquire()  # preventing multipule instances from accessing all at once
        if not path in self or self[path]["stream"].closed:
            self[path] = {
                "stream": open(path, "rb"),
                "read": 0,
                "length": os.stat(path).st_size,
            }
        self.lock.release()

    def close(self, path):
        stream: IOBase = self[path]["stream"]
        stream.close()
        del self[path]

    def read(self, path, start, end):
        if not path in self:
            self.open(path)  # open for new IO handler
        stream: IOBase = self[path]["stream"]
        self.lock.acquire()  # same as open
        stream.seek(start)
        length = end - start
        self[path]["read"] += length
        self.lock.release()
        return stream.read(length)


class FileIterator:
    """__iter__ impl for `FileManager` with i/o usage monitoring"""

    tell = None

    def __init__(self, path, start, end) -> None:
        self.path, self.start, self.end = path, start, end

    def __getattr__(self, name):  # defining fallback
        return {}

    def __iter__(self):
        start = self.start
        for start in range(self.start, self.end, FileManager.CHUNK_SIZE):
            yield file_manager.read(
                self.path, start, min(self.end, start + FileManager.CHUNK_SIZE)
            )
        if start + FileManager.CHUNK_SIZE < self.end:
            yield file_manager.read(self.path, start, self.end)

    def __len__(self):
        return self.end - self.start

    def to_bytes(self):
        buffer = bytearray()
        for chunk in self:
            buffer.extend(chunk)
        return buffer


def get_timestamp() -> int:
    """returns current time in ms"""
    return int(time.time() * 1000)


def check_file(path) -> Tuple[str, str, int]:
    """checks if targeted path is a file then returns its Full Path, Basename, Size (in Bytes)"""
    assert os.path.isfile(path), "%s 不存在，检查视频源配置！（如：yt-dlp配置了时间窗口）" % path
    size = os.stat(path).st_size
    return path, os.path.basename(path), size


chunk_queue = Queue()
file_manager = FileManager()



# -*- coding: utf-8 -*-
class SubmissionVideos(list):
    """Container for all videos within a submission (P-arts)"""

    def extend(self, __iterable) -> None:
        for item in __iterable:
            self.append(item)

    def append(self, video) -> None:
        """Only Submissions or Dict (translated into Submission) will be appended to our list"""
        if isinstance(video, dict):
            # try to interpert it as a list of dictionaries sent by server
            with Submission() as submission:
                submission.title = video.get("title", "")
                submission.video_endpoint = video.get("filename", "")
                submission.video_duration = video.get("duration", 0)
                submission.bvid = video.get("bvid", "")
                submission.biz_id = video.get("cid", "")
                submission.aid = video.get("aid", "")
                submission.stat = video
                submission.parent = self
            return super().append(submission)
        elif isinstance(video, Submission):
            return super().append(video)
        else:
            raise Exception("Either a dict or a Submission object can be supplied.")

    @property
    def archive(self):
        """Dumps current videos as archvies that's to be the payload"""
        target = self if self else [self.parent]  # fallback to parent node
        return [
            {
                "filename": video.video_endpoint,
                "title": video.title,
                "desc": video.description,
                **({"cid": video.biz_id} if video.biz_id else {}),
            }
            for video in target
        ]

    def __init__(self, parent=None):
        """Initializes the list

        parent : Submission - Used as fallback value when theres no subvideos
        """
        self.parent = parent
        super().__init__()

    def __repr__(self) -> str:
        return f"<SubmissionVideos count={len(self)}>"


class Submission:
    """Generic type for a Submission"""

    """COPYRIGHT types"""
    COPYRIGHT_ORIGINAL = 1
    COPYRIGHT_REUPLOAD = 2
    """REPRINT types"""
    REPRINT_DISALLOWED = 1
    REPRINT_ALLOWED = 0
    """Copyright consts"""
    close_reply: bool = False
    close_danmu: bool = False
    """Access control parameters"""
    _description: str = ""

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, v):
        self._description = v or ""  # fallback

    """Description for the video"""
    title: str = ""
    """Title for the submission"""
    copyright: int = COPYRIGHT_REUPLOAD
    """Copyright type"""
    no_reprint: int = REPRINT_ALLOWED
    """Reupload allowance type"""
    source: str = ""
    """Reupload source"""
    thread: int = 0
    """Thread ID"""
    tags: list = None
    """Tags of video"""
    videos: SubmissionVideos = None
    """List of videos in submission"""
    _cover_url = ""

    @property
    def cover_url(self):
        """Cover image URL"""
        return self._cover_url

    @cover_url.setter
    def cover_url(self, value):
        # note : this will strip the HTTP prefix
        if value:
            self._cover_url = "//" + value.split("//")[-1]

    # region Per video attributes
    _video_filename = ""

    @property
    def video_endpoint(self):
        """Endpoint name"""
        return self._video_filename

    @video_endpoint.setter
    def video_endpoint(self, value):
        # note : this will strip the HTTP prefix
        self._video_filename = value.split("/")[-1].split(".")[0]

    biz_id = 0
    """a.k.a cid.for web apis"""
    bvid = ""
    """the new video ID"""
    aid = 0
    """another ID for web apis"""
    thread_name = ""
    """upload thread name i.e. typename"""
    parent_tname = ""
    """parent thread name"""
    stat = None
    """viewer status"""
    reject_reason = ""
    """rejection"""
    state = 0
    """status of video"""
    state_desc = ""
    """status but human readable"""
    video_duration = 0
    """duration of video"""
    desc_format_id = 0
    """description format IDs    
    TODO : Document the IDs with corrosponding Threads"""
    _parent = None

    @property
    def parent(self):
        self._parent: Submission
        return self._parent

    @parent.setter
    def parent(self, v):
        self._parent = v

    """parent object. used for videos property"""
    # endregion
    def __init__(self, title="", desc="", video_endpoint="") -> None:
        self.tags = []  # creates new instance for mutables
        self.videos = SubmissionVideos(self)
        self.title = title
        self.description = desc
        self.video_endpoint = video_endpoint

    def __enter__(self):
        """Creates a new,empty submission"""
        return Submission()

    def __exit__(self, *args):
        pass

    @property
    def archive(self):
        """returns a dict containing all our info"""
        kv_pair = {
            "copyright": self.copyright,
            "videos": self.videos.archive,
            "source": self.source,
            "tid": int(self.thread),
            "title": self.title,
            "tag": ",".join(set(self.tags)),
            "desc_format_id": self.desc_format_id,
            "desc": self.description,
            # "up_close_reply": self.close_reply,
            # "up_close_danmu": self.close_danmu
            "no_reprint": self.no_reprint,
            "cover": self.cover_url
        }
        return kv_pair

    def __repr__(self) -> str:
        return '< bvid : "%s" , thread : %s , title : "%s", desc : "%s" , video_endpoint : "%s" >' % (
            self.bvid,
            self.thread,
            self.title,
            self.description,
            self.video_endpoint
            or ",".join([video.video_endpoint for video in self.videos]),
        )


def create_submission_by_arc(arc: dict):
    """Generates a `Submission` object via a `arc` dict"""
    with Submission() as submission:
        if "parent_tname" in arc:  # Web version only
            submission.parent_tname = arc["parent_tname"]
            submission.thread_name = arc["typename"]
        if "Archive" in arc:
            arc["archive"] = arc["Archive"]
        submission.stat = arc.get("archive",arc.get("stat"))
        submission.aid = submission.stat["aid"]
        submission.thread = submission.stat["tid"]
        submission.desc_format_id = submission.stat["desc_format_id"]
        submission.copyright = arc["archive"]["copyright"]
        submission.bvid = arc["archive"]["bvid"]
        submission.title = arc["archive"]["title"]
        submission.cover_url = arc["archive"]["cover"]
        submission.tags = arc["archive"]["tag"].split(",")
        submission.description = arc["archive"]["desc"]
        submission.source = arc["archive"]["source"]
        submission.state_desc = arc["archive"]["state_desc"]
        submission.state = arc["archive"]["state"]
        submission.reject_reason = arc["archive"]["reject_reason"]
        submission.no_reprint = arc["archive"]["no_reprint"]
        if "Videos" in arc:
            arc["videos"] = arc["Videos"]
        submission.videos.extend(arc["videos"])
    return submission



# -*- coding: utf-8 -*-
"""Youtube video provider - yt-dlp"""
from sqlite3 import Date
from yt_dlp.postprocessor.ffmpeg import FFmpegPostProcessor, FFmpegPostProcessorError
from yt_dlp.postprocessor.embedthumbnail import FFmpegThumbnailsConvertorPP
from yt_dlp.utils import (
    encodeArgument,
    encodeFilename,
    prepend_extension,
    shell_quote,
    DateRange,
    datetime_from_str,
)
from yt_dlp.version import __version__ as yt_dlp_version
from bilibili_toolman.providers import DownloadResult
import logging, yt_dlp, os, subprocess, sys

__desc__ = """Youtube / Twitch / etc 视频下载 (yt-dlp %s)""" % yt_dlp_version
__cfg_help__ = """yt-dlp 参数：
    format (str) - 同 yt-dlp -f
    quite (True,False) - 是否屏蔽 yt-dlp 日志 (默认 False)
特殊参数：
    playlistend - 对于播放列表、频道，下载到（时间顺序，新者在前）第 n 个视频为止
    playliststart - 对于播放列表、频道，从（时间顺序，新者在前）第 n 个视频开始下载

    daterange - 只下载在该参数指定时间窗口内的视频 (精确到毫秒)
        格式可以为 YYmmdd,也可以用相对时间. 如：
        
        e.g. daterange=now; (下载今天上传的视频)
        e.g. daterange=now-1day; (下载昨天到今天上传的视频)
        e.g. daterange=220430~220501 (下载 2022年4月30日~2022年5月1日 的视频)        
    
    hardcode - 烧入硬字幕选项
        e.g. 启用    ..;hardcode;...
        e.g. 换用字体 ..;hardcode=style:FontName=Segoe UI       
        e.g. NV硬解码   ..;hardcode=input:-hwaccel cuda/output:-c:v h264_nvenc -crf 17 -b:v 5M
        多个选项用 / 隔开   
e.g. --youtube "..." --opts "format=best&quiet=True&hardcode" --tags ...
    此外，针对视频对象，还提供其他变量:
        {id}
        {title}    
        {descrption}
        {upload_date}
        {uploader}
        {uploader_id}
        {uploader_url}
        {channel_id}
        {channel_url}
        {duration}
        {view_count}
        {avereage_rating}
        ...
    注：输入播放列表且多 P 时，稿件标题为播放列表标题，稿件描述仅为 `来自Youtube`

默认配置：不烧入字幕，下载最高质量音视频，下载字幕但不操作
"""
ydl = None
logger = logging.getLogger("yt-dlp")
yt_dlp.utils.std_headers[
    "User-Agent"
] = "Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)"
params = {
    "logger": logger,
    "outtmpl": "%(id)s.%(ext)s",
    "format": "bestvideo[ext=mp4]+bestaudio[ext=m4a]/best[ext=mp4]/best",
    "writethumbnail": True,
    "writesubtitles": True,
}  # default params,can be overridden


def __to_yyyy_mm_dd(date):
    return date[:4] + "/" + date[4:6] + "/" + date[6:]


def update_config(cfg):
    global ydl
    # preprocess some parameters
    if "daterange" in cfg:
        datestr = cfg["daterange"]
        daterange = DateRange()
        if "~" in datestr:
            dateStart, dateEnd = datestr.split("~")
            daterange.start = datetime_from_str(dateStart, precision="second").date()
            daterange.end = datetime_from_str(dateEnd, precision="second").date()
        else:
            daterange.start = datetime_from_str(datestr, precision="second").date()
        cfg["daterange"] = daterange
        logger.info("指定要下载的视频上传时间窗口: %s - %s" % (
            daterange.start.strftime("%Y/%m/%d %H:%M:%S"),
            daterange.end.strftime("%Y/%m/%d %H:%M:%S")
        ))

    if "playlistend" in cfg:
        cfg["playlistend"] = int(cfg["playlistend"])
    if "playlistbegin" in cfg:
        cfg["playlistbegin"] = int(cfg["playlistbegin"])

    hardcodeSettings = None
    if "hardcode" in cfg:  # private implementation of hardcoding subtitles
        hardcodeSettings = HardcodeSettings(from_cmd=cfg["hardcode"])
        del cfg["hardcode"]
    ydl = yt_dlp.YoutubeDL({**params, **cfg})
    ydl.add_post_processor(FFmpegThumbnailsConvertorPP(ydl, format="png"))
    if hardcodeSettings:
        ydl.add_post_processor(HardcodeSubProcesser(ydl, hardcodeSettings))


class HardcodeSettings:
    style = "FontName=Segoe UI,FontSize=24"
    """alternative font style for subs filter"""
    input = ""  # params for input file
    output = ""  # params for output file
    """other FFMPEG parameters"""

    def __init__(self, from_cmd):
        """Constructs settings via commandline"""
        for cmd in from_cmd.split("/"):
            if not cmd:
                break
            idx = cmd.index(":")
            key, value = cmd[:idx], cmd[idx + 1 :]
            setattr(self, key, value)


class HardcodeSubProcesser(FFmpegPostProcessor):
    def run_ffmpeg_multiple_files(self, input_paths, out_path, opts):
        """making ffmpeg output to stdout instead,and allowing input parameters"""
        self.check_version()

        oldest_mtime = min(
            os.stat(encodeFilename(path)).st_mtime for path in input_paths
        )

        opts += self._configuration_args()

        files_cmd = []
        for path in input_paths:
            files_cmd.extend(
                [
                    encodeArgument("-i"),
                    encodeFilename(self._ffmpeg_filename_argument(path), True),
                ]
            )
        cmd = [
            encodeFilename(self.executable, True),
            encodeArgument("-y"),
            encodeArgument("-hide_banner"),
        ] + self.settings.input.split()
        # avconv does not have repeat option
        if self.basename == "ffmpeg":
            cmd += [
                encodeArgument("-loglevel"),
                encodeArgument("warning"),
                encodeArgument("-stats"),
            ]
        cmd += (
            files_cmd
            + [encodeArgument(o) for o in opts]
            + [encodeFilename(self._ffmpeg_filename_argument(out_path), True)]
        )
        self._downloader.to_screen("[debug] ffmpeg command line: %s" % shell_quote(cmd))
        p = subprocess.Popen(
            cmd, stdout=sys.stdout, stderr=sys.stderr, stdin=subprocess.PIPE
        )
        p.communicate()
        if p.returncode != 0:
            raise FFmpegPostProcessorError("See stderr for more info")
        self.try_utime(out_path, oldest_mtime, oldest_mtime)

    def __init__(self, downloader, settings: HardcodeSettings):
        self.settings = settings
        super().__init__(downloader=downloader)

    def run(self, information):
        sub = information["requested_subtitles"]
        if sub:
            lang = list(sub.keys())[0]
            sub_filename = f"{information['display_id']}.{lang}.vtt"
            if os.path.isfile(sub_filename):
                self._downloader.to_screen("[ffmpeg] 烧入字幕: %s" % sub_filename)
                filename = information["filepath"]
                opts = [
                    "-vf",
                    f"subtitles={sub_filename}:force_style='{self.settings.style}'",
                    "-qscale",
                    "0",
                    "-c:a",
                    "copy",
                ] + self.settings.output.split()

                temp_filename = prepend_extension(filename, "temp")
                self.run_ffmpeg(filename, temp_filename, opts)

                os.remove(encodeFilename(filename))
                os.rename(encodeFilename(temp_filename), encodeFilename(filename))
        return [], information  # by default, keep file and do nothing


def download_video(res) -> DownloadResult:
    def format_desc(entry):
        return

    with DownloadResult() as results:
        # downloading the cover
        def append_result(entry):
            with DownloadResult() as result:
                result.extra = entry
                result.title = entry["title"]
                result.soruce = entry["webpage_url"]
                result.video_path = "%s.%s" % (entry["display_id"], entry["ext"])
                """For both total results and local sub-results"""
                results.cover_path = result.cover_path = "%s.%s" % (
                    entry["display_id"],
                    "png",
                )
                if not os.path.isfile(results.cover_path):
                    logger.error("Thumbnail not found. Discarding info.")
                    results.cover_path = ""
                date = __to_yyyy_mm_dd(entry["upload_date"])
                result.description = f"""作者 : {entry['uploader']} [{date} 上传]

来源 : https://youtu.be/{entry['id']}

{entry['description']}"""
            results.results.append(result)

        info = ydl.extract_info(res, download=True)
        results.soruce = info["webpage_url"]
        results.title = info["title"]
        """Appending our results"""
        if "entries" in info:  # A playlist
            for entry in info["entries"]:
                append_result(entry)
            results.description = "转自Youtube"
        else:  # Singular videos
            append_result(info)        
        if len(results.results) > 1:
            results.description = '\n'.join(['P%d : %s' % result for result in enumerate(results.results)])
        else:
            results.description = results.results[0].description
        return results
